@@ -68,7 +68,7 @@
 //!     }
 //! }
 //!
-//! pub fn start() -> Result<(), ExitError> {
+//! pub fn start() -> Result<(), ExitError<&'static str>> {
 //!     let init_file = open("/sbin/init", OFlag::O_RDONLY, Mode::empty())
 //!         .printable(PROGRAM_NAME, "unable to open /sbin/init")
 //!         .bail(1)?;
@@ -102,15 +102,15 @@ use std::{
     error::Error,
     ffi::c_void,
     fmt::{
+        Debug,
         Display,
         Formatter,
     },
-    os::unix::io::RawFd,
-    process::{abort, exit},
+    process::exit,
 };
 use nix::{
     errno::Errno,
-    libc::{_exit, write},
+    libc::{STDERR_FILENO, _exit, write},
 };
 
 /// This is the base struct containing basic error information. Unless
@@ -126,7 +126,7 @@ use nix::{
 /// Printed error tries to follow `perror(3p)`'s format with the
 /// addition of the program name.
 #[derive(Debug)]
-pub struct PrintableErrno {
+pub struct PrintableErrno<S: AsRef<str>> {
     /// The name of the program responsible for generating this Error.
     /// This is normally the name of the final binary crate, usually
     /// saved as a const in the main.rs file.
@@ -138,16 +138,16 @@ pub struct PrintableErrno {
     /// `open("/path/to/file", OFlag::O_RDONLY, Mode::empty())`, the message could be as
     /// simple as `open`, or more detailed like `unable to open /path/to/file`.
     /// `PrintableErrno` will fill in the description of the [Errno] if it was available.
-    message: String,
+    message: S,
 
     /// The originating [Errno] value. This is always present unless this `PrintableErrno`
     /// originates from a call to [printable_error].
     errno: Option<Errno>,
 }
-impl PrintableErrno {
+impl <S: AsRef<str>> PrintableErrno<S> {
     /// Attach an exit code to the Error. Useful for bubbling up errors from a deep
     /// function using the `?` operator.
-    pub fn bail(self, exit_code: i32) -> ExitError {
+    pub fn bail(self, exit_code: i32) -> ExitError<S> {
         ExitError {
             exit_code,
             errno: self,
@@ -173,51 +173,83 @@ impl PrintableErrno {
     /// first write. However, any subsequent writes may silently fail if stderr is closed
     /// while this function is running.
     pub fn eprint_signalsafe(&self) {
-        const STDERR: RawFd = nix::libc::STDERR_FILENO;
+        // XXX: Is writev signal-safe?
+        // signal-safety(7) lists write(2) as signal-safe, but not writev(2).
+        // However writev(2) states:
+        //     The writev() system call works just like write(2) except that
+        //     multiple buffers are written out.
+        // Since write(2) is signal-safe, and writev(2) claims to work like
+        // write(2), is its omission from signal-safety(7) an oversight? Or
+        // is there some other undocumented, unexplained difference between
+        // them? Furthermore, since writev(2)'s signal-safety is unclear, can
+        // we rely on POSIX-compliant OSes and libc/library authors to not
+        // introduce any point of unsafety in this regard? i.e. even if the
+        // CSRG issues a revision to the POSIX standard tomorrow that explicitly
+        // requires writev(2) to be signal-safe, can we be sure that all
+        // supported Linux versions comply, as well as all supported glibc/musl
+        // versions in use?
+        // Further reading:
+        // https://groups.google.com/g/comp.unix.programmer/c/IfHr5I8NwW0
+        // https://pubs.opengroup.org/onlinepubs/9699919799/functions/V2_chap02.html
+        // https://austingroupbugs.net/view.php?id=1455
         const CONST_COLON: &'static [u8] = b": ";
         const CONST_NEWLINE: &'static [u8] = b"\n";
 
         let program_name = &self.program_name.as_bytes();
-        let message = &self.message.as_bytes();
-        let res = unsafe { write(STDERR, program_name.as_ptr() as *const c_void, program_name.len()) };
-        if res < 0 {
-            // abort() should be signal-safe according to ISO C 99 (7.14.1.1p5).
-            // However, some versions of POSIX (before 1003.1-2004) required abort()
-            // to flush stdio streams, which would be hard (or impossible) to do
-            // safely while upholding the contract. In fact there are glibc
-            // implementations (before commit 91e7cf982d01) that perform the flush
-            // unsafely.
-            // Fortunately, Rust's stdlib makes no use of stdio buffering, so we
-            // can assume that most unsafe abort() implementations will be safe
-            // for us to call, as no buffer will be flushed.
-            abort()
+        let message = &self.message.as_ref().as_bytes();
+        let res = write_stderr(program_name);
+        if !res {
+            // error writing to stderr: there is nothing left to do
+            return;
         }
 
-        unsafe {
-            write(STDERR, CONST_COLON.as_ptr() as *const c_void, CONST_COLON.len());
-            write(STDERR, message.as_ptr() as *const c_void, message.len());
-        }
+        write_stderr(CONST_COLON);
+        write_stderr(message);
 
         let errno = self.errno.as_ref();
         if let Some(errno) = errno {
             let desc = errno.desc().as_bytes();
-            unsafe {
-                write(STDERR, CONST_COLON.as_ptr() as *const c_void, CONST_COLON.len());
-                write(STDERR, desc.as_ptr() as *const c_void, desc.len());
-                write(STDERR, CONST_NEWLINE.as_ptr() as *const c_void, CONST_NEWLINE.len());
+            write_stderr(CONST_COLON);
+            write_stderr(desc);
+        }
+        write_stderr(CONST_NEWLINE);
+
+        fn write_stderr(buf: &[u8]) -> bool {
+            let mut pos = buf.as_ptr() as usize;
+            let mut len = buf.len();
+            loop {
+                let res = unsafe { write(STDERR_FILENO, pos as *const c_void, len) };
+                match res {
+                    0 => break true,
+                    e if e < 0 => {
+                        let err = Errno::last();
+                        if !matches!(err, Errno::EINTR) {
+                            break false;
+                        }
+                        // EINTR means we can try again, so continue loop
+                    }
+                    // written will always be positive
+                    w if (w as usize) < len => {
+                        let written = w as usize;
+                        pos += written;
+                        len -= written;
+                        // try again with remaining
+                    }
+                    _ => break true,
+                }
             }
         }
     }
 }
-impl Display for PrintableErrno {
+impl <S: AsRef<str>> Display for PrintableErrno<S> {
     fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
         match self.errno {
-            Some(errno) => write!(f, "{}: {}: {}", self.program_name, self.message, errno.desc()),
-            None => write!(f, "{}: {}", self.program_name, self.message),
+            Some(errno) => write!(f, "{}: {}: {}", self.program_name, self.message.as_ref(), errno.desc()),
+            None => write!(f, "{}: {}", self.program_name, self.message.as_ref()),
         }
     }
 }
-impl Error for PrintableErrno {
+impl <S: AsRef<str> + Debug> Error for PrintableErrno<S> {
     fn source(&self) -> Option<&(dyn Error + 'static)> {
         match self.errno {
             Some(ref x) => Some(x),
@@ -245,11 +277,11 @@ impl Error for PrintableErrno {
 /// and `_exit(exit_code)` respectively, exiting the process with the supplied
 /// exit code.
 #[derive(Debug)]
-pub struct ExitError {
+pub struct ExitError<S: AsRef<str>> {
     exit_code: i32,
-    errno: PrintableErrno,
+    errno: PrintableErrno<S>,
 }
-impl ExitError {
+impl <S: AsRef<str>> ExitError<S> {
     /// Print the error to stderr and exit with the supplied code.
     ///
     /// # Safety
@@ -278,14 +310,14 @@ impl ExitError {
         }
     }
 }
-impl Display for ExitError {
+impl <S: AsRef<str>> Display for ExitError<S> {
     fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
         self.errno.fmt(f)
     }
 }
-impl Error for ExitError {
+impl <S: AsRef<str> + Debug> Error for ExitError<S> {
     fn source(&self) -> Option<&(dyn Error + 'static)> {
-        Some(&self.errno)
+        self.errno.source()
     }
 }
 
@@ -297,10 +329,10 @@ pub trait ErrnoResult<T> {
     /// Maps the nix [Errno] to a [PrintableErrno] with the program name and the
     /// supplied message. This allows [nix::Result]s to be easily used with this
     /// library's error handling and Rust's `?` operator.
-    fn printable<S: Into<String>>(self, program_name: &'static str, message: S) -> Result<T, PrintableErrno>;
+    fn printable<S: AsRef<str>>(self, program_name: &'static str, message: S) -> Result<T, PrintableErrno<S>>;
 }
 impl <T> ErrnoResult<T> for Result<T, Errno> {
-    fn printable<S: Into<String>>(self, program_name: &'static str, message: S) -> Result<T, PrintableErrno> {
+    fn printable<S: AsRef<str>>(self, program_name: &'static str, message: S) -> Result<T, PrintableErrno<S>> {
         self.map_err(|e| PrintableErrno {
             program_name,
             message: message.into(),
@@ -313,10 +345,10 @@ impl <T> ErrnoResult<T> for Result<T, Errno> {
 
 /// This trait defines a [Result] with our own [PrintableErrno], allowing further conversion
 /// to [ExitError], or printing and returning and Option.
-pub trait PrintableResult<T> {
+pub trait PrintableResult<T, S: AsRef<str>> {
     /// Maps the [PrintableErrno] to an [ExitError] in order to facilitate
     /// bubbling up the error to the appropriate place to quit the program.
-    fn bail(self, exit_code: i32) -> Result<T, ExitError>;
+    fn bail(self, exit_code: i32) -> Result<T, ExitError<S>>;
 
     /// Consumes the result, prints the error (if present) to stderr, and returns
     /// an [Option] of the underlying value.
@@ -333,8 +365,8 @@ pub trait PrintableResult<T> {
     /// The caller must ensure that stderr (fd 2) is open.
     fn ok_or_eprint_signal_safe(self) -> Option<T>;
 }
-impl <T> PrintableResult<T> for Result<T, PrintableErrno> {
-    fn bail(self, exit_code: i32) -> Result<T, ExitError> {
+impl <T, S: AsRef<str>> PrintableResult<T, S> for Result<T, PrintableErrno<S>> {
+    fn bail(self, exit_code: i32) -> Result<T, ExitError<S>> {
         self.map_err(|e| e.bail(exit_code))
     }
     fn ok_or_eprint(self) -> Option<T> {
@@ -406,7 +438,7 @@ pub trait ExitErrorResult<T> {
     /// The caller must ensure that stderr (fd 2) is open.
     fn ok_or_eprint_signal_safe(self) -> Option<T>;
 }
-impl <T> ExitErrorResult<T> for Result<T, ExitError> {
+impl <T, S: AsRef<str>> ExitErrorResult<T> for Result<T, ExitError<S>> {
     fn unwrap_or_eprint_exit(self) -> T {
         match self {
             Ok(t) => t,
@@ -448,7 +480,7 @@ impl <T> ExitErrorResult<T> for Result<T, ExitError> {
 /// Creates an error out of the program name and the supplied message. Useful
 /// when an Errno isn't present, but the program wants to signal a problem to
 /// the user and possibly exit if it's fatal.
-pub fn printable_error<S: Into<String>>(program_name: &'static str, message: S) -> PrintableErrno {
+pub fn printable_error<S: AsRef<str>>(program_name: &'static str, message: S) -> PrintableErrno<S> {
     PrintableErrno {
         program_name,
         message: message.into(),
@@ -464,8 +496,6 @@ mod tests {
     use nix::fcntl::{open, OFlag};
     use nix::sys::stat::Mode;
     use crate::{ErrnoResult, PrintableResult, printable_error};
-    use nix::unistd::{fork, ForkResult};
-    use nix::sys::wait::{waitpid, WaitStatus};
 
     const TEST_NAME: &str = "precisej-printable-errno";
 
@@ -474,7 +504,7 @@ mod tests {
         println!();
         println!("START TEST 1");
 
-        const MSG1_NAME: &str = "unable to open /pathto/nonexistent/file";
+        const MSG1_NAME: &str = "test1: unable to open /pathto/nonexistent/file";
 
         let result1 = open("/pathto/nonexistent/file", OFlag::O_RDONLY, Mode::empty())
             .printable(TEST_NAME, MSG1_NAME);
@@ -496,7 +526,7 @@ mod tests {
         println!();
         println!("START TEST 2");
 
-        const MSG2_NAME: &str = "expected value 1, got 30";
+        const MSG2_NAME: &str = "test2: expected value 1, got 30";
 
         let result1 = printable_error(TEST_NAME, MSG2_NAME);
         let result2 = printable_error(TEST_NAME, MSG2_NAME);
@@ -515,7 +545,7 @@ mod tests {
         println!();
         println!("START TEST 3");
 
-        const MSG3_NAME: &str = "unable to open /pathto/nonexistent/file";
+        const MSG3_NAME: &str = "test3: unable to open /pathto/nonexistent/file";
         const MSG3_EXIT_CODE: i32 = 1;
 
         let result1 = open("/pathto/nonexistent/file", OFlag::O_RDONLY, Mode::empty())
@@ -538,36 +568,21 @@ mod tests {
     }
 
     #[test]
-    fn bail_signal_safe() {
+    fn eprint_signal_safe() {
         println!();
         println!("START TEST 4");
 
-        const MSG4_NAME: &str = "SIGNAL_SAFE: unable to open /pathto/nonexistent/file: No such file or directory";
-        const MSG4_EXIT_CODE: i32 = 1;
+        const MSG4_NAME_1: &str = "test4: unable to open /pathto/nonexistent/file";
+        const MSG4_NAME_2: &str = "test4: expected value 1, got 30";
 
-        match unsafe { fork() }.unwrap() {
-            ForkResult::Parent { child } => {
-                loop {
-                    let waitpid = waitpid(child, None).unwrap();
-                    match waitpid {
-                        WaitStatus::Exited(_, code) => {
-                            println!("CODE: {}", code);
-                            assert_eq!(code, MSG4_EXIT_CODE);
-                            break
-                        }
-                        WaitStatus::Signaled(_, _, _) => {
-                            panic!("Child Signaled, expected Exited.")
-                        }
-                        _ => {
-                            // Continue
-                        }
-                    }
-                }
-            }
-            ForkResult::Child => unsafe {
-                printable_error(TEST_NAME, MSG4_NAME).bail(MSG4_EXIT_CODE).eprint_signal_safe_exit()
-            }
-        }
+        let result1 = open("/pathto/nonexistent/file", OFlag::O_RDONLY, Mode::empty())
+            .printable(TEST_NAME, MSG4_NAME_1);
+        result1.ok_or_eprint_signal_safe();
+
+        let result2 = printable_error(TEST_NAME, MSG4_NAME_2);
+        let result3 = printable_error(TEST_NAME, MSG4_NAME_2);
+        result2.eprint_signalsafe();
+        result3.eprint_signalsafe();
 
         println!("END TEST 4");
     }
